#ifndef __QUANT_OCTREE_H__
#define __QUANT_OCTREE_H__

#include "QuantTypes.h"

int quantize_octree(Pixel *,
          uint32_t,
          uint32_t,
          Pixel **,
          uint32_t *,
          uint32_t **,
          int);

#endif
